'''Core features include an init function for skeleton'''
import re
import logging
from typing import Dict, Optional, Any
import click
from munch import Munch, DefaultFactoryMunch  # type: ignore
from click_help_colors.utils import _colorize  # type: ignore
from click_skeleton.advanced_group import AdvancedGroup
from click_skeleton.decorators import add_options
from click_skeleton.version import version_cmd
from click_skeleton.completion import completion_cli
from click_skeleton.defaults import DEFAULT_CONTEXT_SETTINGS

logger = logging.getLogger(__name__)


def version_option(
        version: Optional[str] = None,
        prog_name: Optional[str] = None,
        message: str = "%(prog)s, version %(version)s",
        message_color: Optional[str] = None,
        prog_name_color: Optional[str] = None,
        version_color: Optional[str] = None,
        **kwargs: Any,
) -> Any:
    '''Re-implement version handling with --version and -V shortcut'''
    msg_parts = []
    for placeholder in re.split(r'(%\(version\)s|%\(prog\)s)', message):
        if placeholder == '%(prog)s':
            msg_parts.append(_colorize(prog_name, prog_name_color or message_color))
        elif placeholder == '%(version)s':
            msg_parts.append(_colorize(version, version_color or message_color))
        else:
            msg_parts.append(_colorize(placeholder, message_color))
    message = ''.join(msg_parts)

    return click.version_option(
        version,
        '--version', '-V',
        prog_name=prog_name,
        message=message,
        **kwargs
    )


def skeleton(
    name: str,
    version: str,
    auto_envvar_prefix: Optional[str] = None,
    cls: Any = None,
    commands: Optional[Dict[str, Any]] = None,
    **kwargs: Any,
) -> Any:
    '''Generates an skeleton group with version options included'''
    auto_envvar_prefix = auto_envvar_prefix if auto_envvar_prefix is not None else name.upper()
    if cls is None:
        cls = AdvancedGroup

    sensible_context_settings = DEFAULT_CONTEXT_SETTINGS
    sensible_context_settings['auto_envvar_prefix'] = auto_envvar_prefix

    obj = DefaultFactoryMunch(Munch)
    obj.prog_name = name
    obj.version = version
    obj.context_settings = sensible_context_settings

    context_settings = sensible_context_settings
    context_settings['auto_envvar_prefix'] = auto_envvar_prefix
    context_settings['obj'] = obj

    commands = commands if commands is not None else {}
    commands['completion'] = completion_cli
    commands['version'] = version_cmd
    return add_options(
        version_option(
            version=version,
            prog_name=name,
            version_color='green',
            prog_name_color='yellow',
        ),
        click.group(
            name=name,
            context_settings=context_settings,
            commands=commands,
            cls=cls,
            **kwargs,
        ),
    )



'''Generates a readme composed of all commands and subcommands help strings'''
from textwrap import indent
from typing import Dict, Any
import click
from . import AdvancedGroup


def write_header(text: str, char: str) -> None:
    '''Write line followed by a newline containing as many chars'''
    click.echo("\n" + text)
    click.echo(char * len(text))


def write_codeblock(code: str, output: str) -> None:
    '''Write a console codeblock, either in rst or markdown'''
    if output == 'rst':
        indented_code = indent(code, '  ')
        click.echo(".. code-block::\n")
        click.echo(indented_code)
    elif output == 'markdown':
        click.echo('```')
        click.echo(code)
        click.echo('```')


def readme(
    main_cli: click.core.Group,
    prog_name: str,
    context_settings: Dict[str, Any],
    output: str = 'rst',
) -> Any:
    '''Recursively output a beautiful readme on stdout, supports 3 levels of subcommands'''
    if output == 'rst':
        write_header("Commands", '-')
    elif output == 'markdown':
        click.echo('# Commands\n')

    base_ctx = click.Context(main_cli, info_name=prog_name, **context_settings)
    with base_ctx.scope():
        cli_help = main_cli.get_help(base_ctx)
        write_codeblock(cli_help, output)

        for command_name, command in sorted(main_cli.commands.items()):
            if command.hidden:
                continue

            command_header = f"{prog_name} {command_name}"
            if output == 'rst':
                write_header(command_header, '*')
            elif output == 'markdown':
                click.echo(f'\n## {command_header}\n')

            command_ctx = click.Context(command, info_name=command_name, parent=base_ctx)
            with command_ctx.scope():
                command_help = command.get_help(command_ctx)
                write_codeblock(command_help, output)

            if not isinstance(command, AdvancedGroup):
                continue

            for subcommand_name, subcommand in sorted(command.commands.items()):
                if subcommand_name == 'help' or subcommand.hidden:
                    continue

                subcommand_header = f"{prog_name} {command_name} {subcommand_name}"
                if output == 'rst':
                    write_header(f"{subcommand_header}", '*')
                elif output == 'markdown':
                    click.echo(f'\n### {subcommand_header}\n')

                subcommand_ctx = click.Context(subcommand, info_name=subcommand_name, parent=command_ctx)
                with subcommand_ctx.scope():
                    subcommand_help = subcommand.get_help(subcommand_ctx)
                    write_codeblock(subcommand_help, output)



'''Replace default stacktrace printing with a beautiful one'''
import os
import sys
import traceback
from typing import Any, Union, List, Tuple, Optional
from types import TracebackType
import click


TRACEBACK_IDENTIFIER = 'Traceback (most recent call last):\n'


def _flush(message: str) -> None:
    '''Print message and flush stderr'''
    sys.stderr.write(message + '\n')
    sys.stderr.flush()


class _Hook:
    '''Stacktrace hook'''
    def __init__(
            self,
            entries: Any,
            strip_path: bool = False,
    ):
        self.entries = entries
        self.strip = strip_path

    def reverse(self) -> None:
        '''Reverse stacktrace entries to ease reading'''
        self.entries = self.entries[::-1]

    def rebuild_entry(self, entry: Any) -> Any:
        '''Rebuild context of entry'''
        entry = list(entry)
        # This is the file path.
        entry[0] = os.path.basename(entry[0]) if self.strip else entry[0]
        # Always an int (entry line number)
        entry[1] = str(entry[1])

        new_entry = [
            click.style(f"{entry[1]}", fg="bright_red"),
            entry[0],
            click.style(f"{entry[2]}", fg="bright_green"),
            click.style(f"--> {entry[3]}", fg="bright_yellow"),
        ]

        return new_entry

    @staticmethod
    def align_all(entries: Any) -> List[int]:
        '''Align all stacktrace entries'''
        lengths = [0, 0, 0, 0]
        for entry in entries:
            for index, field in enumerate(entry):
                lengths[index] = max(lengths[index], len(str(field)))
        return lengths

    @staticmethod
    def align_entry(entry: Any, lengths: List[int]) -> str:
        '''Align one stacktrace entry'''
        return ' '.join(
            ['{0:{1}}'.format(field, lengths[index])  # pylint: disable=consider-using-f-string
             for index, field in enumerate(entry)]
        )

    def generate_backtrace(self) -> List[str]:
        '''Return the (potentially) aligned, rebuit traceback

        Yes, we iterate over the entries thrice. We sacrifice
        performance for code readability. I mean.. come on, how long can
        your traceback be that it matters?
        '''
        backtrace = []
        for entry in self.entries:
            backtrace.append(self.rebuild_entry(entry))

        # Get the lenght of the longest string for each field of an entry
        lengths = self.align_all(backtrace)
        return [self.align_entry(entry, lengths) for entry in backtrace]


def hook(
    strip_path: bool = False,
    enable_on_envvar_only: bool = False,
    on_tty: bool = False,
    trace: Optional[TracebackType] = None,
    tpe: Optional[type] = None,
    value: Optional[BaseException] = None,
) -> None:
    '''Hook'''
    if enable_on_envvar_only and 'ENABLE_BACKTRACE' not in os.environ:
        return

    isatty = getattr(sys.stderr, 'isatty', lambda: False)
    if on_tty and not isatty():
        return

    def backtrace_excepthook(
        tpe: Optional[Union[str, type]],
        value: Optional[BaseException],
        trace: Optional[TracebackType] = None,
    ) -> None:
        '''Don't know if we're getting traceback or traceback entries.
        We'll try to parse a traceback object.
        '''
        try:
            traceback_entries = traceback.extract_tb(trace)
            parser = _Hook(traceback_entries, strip_path)
        except AttributeError:
            parser = _Hook(trace, strip_path)

        if tpe is None:
            type_str = 'unknown'
        elif isinstance(tpe, str):
            type_str = tpe
        else:
            type_str = tpe.__name__

        tb_message = click.style('Traceback (Most recent call last):', fg="yellow")
        err_message = click.style(f"{type_str}: {value}", fg="bright_red")

        _flush(tb_message)
        backtrace = parser.generate_backtrace()
        backtrace.insert(len(backtrace), err_message)
        for entry in backtrace:
            _flush(entry)

    if trace:
        backtrace_excepthook(tpe=tpe, value=value, trace=trace)
    else:
        sys.excepthook = backtrace_excepthook


def unhook() -> None:
    '''Restore the default excepthook'''
    sys.excepthook = sys.__excepthook__  # type: ignore


def _extract_traceback(text: str) -> Tuple[List[Tuple[str, str, str]], List[str]]:
    '''Receive a list of strings representing the input from stdin and return
    the restructured backtrace.

    This iterates over the output and once it identifies a hopefully genuine
    identifier, it will start parsing output.
    In the case the input includes a reraise (a Python 3 case), the primary
    traceback isn't handled, only the reraise.

    Each of the traceback lines are then handled two lines at a time for each
    stack object.

    Note that all parts of each stack object are stripped from newlines and
    spaces to keep the output clean.
    '''
    capture = False
    entries = []
    all_else = []
    ignore_trace = False

    # In python 3, a traceback may includes output from a reraise.
    # e.g, an exception is captured and reraised with another exception.
    # This marks that we should ignore
    if text.count(TRACEBACK_IDENTIFIER) == 2:
        ignore_trace = True

    for line in text:
        if TRACEBACK_IDENTIFIER in line:
            if ignore_trace:
                ignore_trace = False
                continue
            capture = True
            location_info = True
        # We're not capturing and making sure we only read lines
        # with spaces since, after the initial identifier, all traceback lines
        # contain a prefix spacing.
        elif capture and line.startswith(' '):
            if location_info:
                # Line containing a file, line and module.
                line = line.strip()
                entries.append(line)
            else:
                # The corresponding line of source code.
                line = line.rstrip("\n")
                entries[-1] += ', ' + line
            location_info = not location_info
        elif capture:
            # Line containing the module call.
            entries.append(line)
            break
        else:
            # Add everything else before the traceback.
            all_else.append(line)

    traceback_entries = []
    # Build the traceback structure later passed for formatting.
    for _, line in enumerate(entries[:-2]):
        element = line.split(',', 3)
        _file = element[0].strip().lstrip('File').strip(' "')
        _line = element[1].strip().lstrip('line').strip()
        _in = element[2].strip().lstrip('in').strip()
        traceback_entries.append((_file, _line, _in))
    return traceback_entries, all_else



'''Advanced group is a colored, what-did-you-mean with aliases commands group'''
import logging
import pkgutil
import sys
import importlib
from types import ModuleType
from typing import Any, Optional
import click
from click_help_colors import HelpColorsGroup  # type: ignore
from click_didyoumean import DYMGroup  # type: ignore
from click_aliases import ClickAliasedGroup  # type: ignore
from click_skeleton.exceptions import AlreadyRegistered

logger = logging.getLogger(__name__)


class AdvancedGroup(ClickAliasedGroup, DYMGroup, HelpColorsGroup):  # type: ignore
    '''Special click group with default plugins enabled :
    - did-you-mean
    - click aliases for commands
    - colored help message
    - auto help command
    '''
    def __init__(self, *args: Any, aliases: Optional[str] = None, **kwargs: Any):
        kwargs['help_headers_color'] = 'yellow'
        kwargs['help_options_color'] = 'green'
        super().__init__(*args, **kwargs)
        self.aliases = aliases if aliases is not None else []

        @click.command('help')
        @click.pass_context
        def _help(ctx: Optional[click.Context]) -> None:
            '''Print help'''
            # we accept many commands, but only show help of the first one
            if not ctx:
                raise RuntimeError('no click context available')
            if not ctx.parent:
                raise RuntimeError('no click context parent available')
            click.echo(ctx.parent.get_help(), color=ctx.color)
            ctx.exit()

        self.add_command(_help, 'help')

    def add_group(self, cmd: "AdvancedGroup", name: str) -> None:
        """Registers another :class:`Group` with this group.  If the name
        is not provided, the name of the group is used.
        """
        self.add_command(cmd, name)

        if not cmd.aliases:
            return

        if name in self._commands:
            raise AlreadyRegistered(f'{name} group is already registered')

        self._commands[name] = cmd.aliases
        for alias in cmd.aliases:
            if alias in self._aliases:
                raise AlreadyRegistered(f'Alias {alias} is already used by {self._aliases[alias]}')
            self._aliases[alias] = name

    def add_groups_from_package(self, package: ModuleType) -> None:
        '''loads commands dynamically, not supported by pyinstaller'''
        package_name = package.__name__
        package = sys.modules[package_name]
        modules_by_name = {
            name: importlib.import_module(package_name + '.' + name)
            for loader, name, is_pkg in pkgutil.walk_packages(package.__path__)  # type: ignore
        }
        for module_name, module in modules_by_name.items():
            try:
                cli = getattr(module, 'cli')
                if cli.name != 'cli':
                    self.add_group(cli, cli.name)
                else:
                    self.add_group(cli, module_name)
            except AttributeError:
                click.secho(f'''Command module {module_name} does not contain a 'cli' definition''', fg='red', err=True)



'''Decorators and helpers to add options to groups and commands, and compose decorators'''
import collections
from typing import Union, Any, Iterable, Optional
import click


def flatten(iterables: Union[Any, Iterable[Any]]) -> Iterable[Any]:
    '''Recursively flatten argument'''
    for element in iterables:
        if isinstance(element, collections.abc.Iterable) and not isinstance(element, (str, bytes)):
            yield from flatten(element)
        else:
            yield element


def add_options(*options: Any) -> Any:
    '''Helps to add options to command in the form of list'''
    def _add_options(func: Any) -> Any:
        for option in reversed(list(flatten(options))):
            func = option(func)
        return func
    return _add_options


def group(name: Optional[str] = None, **attrs: Any) -> Any:
    """Creates a new :class:`Group` with a function as callback.  This
    works otherwise the same as :func:`command` just that the `cls`
    parameter is set to :class:`Group`.
    """
    from click_skeleton.advanced_group import AdvancedGroup
    attrs.setdefault("cls", AdvancedGroup)
    return click.command(name, **attrs)



'''Generates shell code completion'''
import os
from typing import Optional
import click
import click_completion  # type: ignore
from click_skeleton.advanced_group import AdvancedGroup
from click_skeleton.decorators import add_options


completion_options = [
    click.option(
        '-i', '--case-insensitive',
        help="Case insensitive completion",
        is_flag=True,
    ),
    click.argument(
        'shell',
        required=False,
        type=click_completion.DocumentedChoice(click_completion.core.shells),
    ),
]

append_option = click.option(
    '--append/--overwrite',
    help="Append the completion code to the file",
    is_flag=True,
    default=None,
)


@click.group('completion', short_help="Shell completion", cls=AdvancedGroup)
def completion_cli() -> None:
    '''Shell completion subcommand'''


@completion_cli.command(short_help='Show the click-completion-command completion code', aliases=['print', 'generate'])
@add_options(completion_options)  # type: ignore
def show(shell: str, case_insensitive: bool) -> None:
    '''Generate shell code to enable completion'''
    extra_env = {'_CLICK_COMPLETION_COMMAND_CASE_INSENSITIVE_COMPLETE': 'ON'} if case_insensitive else {}
    click.echo(click_completion.core.get_code(shell, extra_env=extra_env))


@completion_cli.command(short_help='Install the click-completion-command completion')
@add_options(completion_options, append_option)  # type: ignore
@click.argument('path', required=False)
def install(append: bool, case_insensitive: bool, shell: str, path: Optional[str]) -> None:
    '''Auto install shell completion code in your rc file'''
    extra_env = {'_CLICK_COMPLETION_COMMAND_CASE_INSENSITIVE_COMPLETE': 'ON'} if case_insensitive else {}
    shell, path = click_completion.core.install(shell=shell, path=path, append=append, extra_env=extra_env)
    click.echo(f'{shell} completion installed in {path}')


def custom_startswith(string: str, incomplete: str) -> bool:
    '''A custom completion matching that supports case insensitive matching'''
    if os.environ.get('_CLICK_COMPLETION_COMMAND_CASE_INSENSITIVE_COMPLETE'):
        string = string.lower()
        incomplete = incomplete.lower()
    return string.startswith(incomplete)


click_completion.core.startswith = custom_startswith
click_completion.init()



'''Import dynamically all modules in a directory'''
import os
import re
import sys
from typing import Iterator, Dict, Any, List


def import_all_modules(path: str, env: Dict[str, Any]) -> List[str]:
    """ Imports all modules residing directly in directory "path" into the provided environment
        (usually the callers environment). A typical call:
        importdir.do("example_dir", globals())
    """
    absolute_dirname = os.path.dirname(path)
    sys.path.append(absolute_dirname)
    modules = []
    for module_name in sorted(get_module_names_in_dir(absolute_dirname)):
        env[module_name] = __import__(module_name, env)
        if module_name != '__init__':
            modules.append(module_name)
    return modules


def get_module_names_in_dir(path: str) -> Iterator[str]:
    '''Returns a set of all module names residing directly in directory "path"'''
    # File name of a module:
    module_file_regexp = r"(.+)\.py(c?)$"
    # Looks for all python files in the directory (not recursively) and add their name to result:
    for entry in os.listdir(path):
        if os.path.isfile(os.path.join(path, entry)):
            regexp_result = re.search(module_file_regexp, entry)
            if regexp_result:
                yield regexp_result.group(1)



'''Click skeleton helpers to ease CLI definitions'''
import logging
import click
from click_skeleton.advanced_group import AdvancedGroup
from click_skeleton.expanded_path import ExpandedPath
from click_skeleton.core import skeleton
from click_skeleton.decorators import add_options

logger = logging.getLogger(__name__)

# little hacky but prevent click from rewraping
click.formatting.HelpFormatter.write_dl.__defaults__ = (50, 2)  # type: ignore

__all__ = [
    'AdvancedGroup',
    'ExpandedPath',
    'skeleton',
    'add_options',
]



'''Basic helpers used in many CLI'''
import logging
import datetime
import platform
import re
import string
import random
from collections import defaultdict
from typing import List, Any
import click

logger = logging.getLogger(__name__)
true_values = ('enabled', 'y', 'yes', 't', 'true', 'True', 'on', '1')
false_values = ('', 'none', 'disabled', 'n', 'no', 'f', 'false', 'False', 'off', '0')


def split_arguments(ctx, param, value):
    '''Arguments can be comma separated'''
    ctx.params[param.name] = []
    for maybe_argument_list in value:
        for argument in mysplit(maybe_argument_list, ','):
            if argument in ctx.params[param.name]:
                click.secho(f'{argument} was provided too many times', fg='yellow', err=True)
                continue
            ctx.params[param.name].append(clean(argument))
    return ctx.params[param.name]


def clean(text: str):
    '''Clean argument'''
    return text.replace('"', '').replace("'", '')


def recursive_str(data: Any) -> Any:
    '''Recursively convert multiple data type to string'''
    if isinstance(data, str):
        return data
    if isinstance(data, list):
        return [recursive_str(x) for x in data]
    if isinstance(data, dict):
        return {recursive_str(key): recursive_str(val) for key, val in data.items()}
    return str(data)


def str2bool(val: Any) -> bool:
    '''Converts any value to string and detects if it looks like a known bool value'''
    val = str(val).lower()
    if val in true_values:
        return True
    if val in false_values:
        return False
    raise ValueError(f"invalid truth value {val}")


def str_is_true(text: str) -> bool:
    '''Detects a known truthful string'''
    return str(text).lower() in true_values


def str_is_false(text: str) -> bool:
    '''Detects a known falsy string'''
    return str(text).lower() in false_values


def seconds_to_human(seconds: int) -> str:
    '''Human readable duration from seconds'''
    return str(datetime.timedelta(seconds=seconds))


def strip_colors(text: str) -> str:
    '''Remove all ANSI control characters'''
    ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')
    return ansi_escape.sub('', text)


def mysplit(text: str, delimiter: str = ',') -> List[str]:
    '''Split a string without returning empty list if string is empty'''
    return [x for x in text.split(delimiter) if x]


def raise_limits() -> bool:
    '''Permits to open a lot of system handles, supports Windows'''
    if platform.system() == 'Windows':
        logger.warning('Cannot raise system limits on Windows')
        return False
    # resource module is only available on UNIX systems
    import resource  # pylint: disable=import-error
    try:
        soft, hard = resource.getrlimit(resource.RLIMIT_NOFILE)
        logger.info(f"Current limits, soft and hard : {soft} {hard}")
        resource.setrlimit(resource.RLIMIT_NOFILE, (hard, hard))
        return True
    except (ValueError, OSError) as error:
        if 'macOS' in platform.platform():
            logger.info(f"Cannot raise system limits for current user on : {platform.platform()}")
        else:
            logger.critical(f'You may need to check ulimit parameter: {error}')
        return False


def random_password(size: int = 8) -> str:
    '''Generates a random string from size, composed of ascii letters and digits'''
    alphabet = string.ascii_letters + string.digits
    return ''.join(random.choice(alphabet) for i in range(size))


class PrettyDefaultDict(defaultdict):  # type: ignore
    '''Beautiful (correct printing) dict using collections.defaultdict'''
    __repr__ = dict.__repr__



'''Checks if a new version of current program is available on PyPI'''
import logging
import threading
import re
from html.parser import HTMLParser
from typing import Any, Optional, Tuple, List
import click
import requests
import semver  # type: ignore


logger = logging.getLogger(__name__)
DEFAULT_PYPI = 'pypi.org'


class MyParser(HTMLParser):
    '''Parser to extract http links'''
    def __init__(
        self,
        *args: Any,
        output_list: Optional[List[Any]] = None,
        **kwargs: Any
    ):
        super().__init__(*args, **kwargs)
        self.output_list = output_list if output_list is not None else []

    def handle_starttag(
        self,
        tag: str,
        attrs: List[Tuple[str, Optional[str]]]
    ) -> None:
        '''We parse only http links'''
        if tag == 'a':
            self.output_list.append(dict(attrs).get('href'))

    def error(self, message: str) -> None:
        '''Print parsing error, implement parser abstract method'''
        logger.error(message)


class VersionCheckerThread(threading.Thread):
    '''Background thread to check version, to start at beginning of CLI'''
    def __init__(self, prog_name: str, current_version: str, domain: str = DEFAULT_PYPI, autostart: bool = True, **kwargs: Any):
        super().__init__(**kwargs)
        self.prog_name = prog_name
        self.new_version_warning: Optional[str] = None
        self.domain = domain
        self.current_version = current_version
        self.url = f'https://{domain}/simple/{prog_name}/'
        if autostart:
            self.start()

    def run(self) -> None:
        '''This threads auto-start by default and store a result you can read at end of program execution'''
        try:
            resp = requests.get(self.url)
            if not resp.ok:
                logger.info(f'{self.prog_name} : unable to fetch {self.url} : {resp.text}')  # pylint: disable=logging-fstring-interpolation
                return

            parser = MyParser()
            parser.feed(resp.text)
            if not parser.output_list:
                logger.info(f'{self.prog_name} : no packages links detected in {resp.text}')  # pylint: disable=logging-fstring-interpolation
                return

            last_link = parser.output_list[-1]
            last_version_matches = re.search(r'(?:(\d+\.[.\d]*\d+))', last_link)
            if not last_version_matches:
                logger.info(f'{self.prog_name} : no version found in string {last_link}')  # pylint: disable=logging-fstring-interpolation
                return

            last_version = last_version_matches.group(1)

            last_version_info = semver.VersionInfo.parse(last_version)
            current_version_info = semver.VersionInfo.parse(self.current_version)
            if current_version_info < last_version_info:
                pip_extra_index_url = ''
                pipx_extra_index_url = ''
                if self.domain != DEFAULT_PYPI:
                    pip_extra_index_url = f'--extra-index-url=https://{self.domain} '
                    pipx_extra_index_url = f'--index-url=https://{self.domain} '
                self.new_version_warning = click.style(
                    f'''
{self.prog_name} : new version {last_version} available (current version: {self.current_version})
upgrade command :
    pip3 install -U {pip_extra_index_url}{self.prog_name}
    pipx upgrade {pipx_extra_index_url}{self.prog_name}''',
                    fg='bright_blue',
                )
        except Exception as error:  # pylint: disable=broad-except
            logger.exception(error)

    def print(self) -> None:
        '''Join thread and print result on stderr'''
        try:
            if not self.is_alive():
                logger.debug('Version checker thread may have exited')
                return
            self.join()
            if self.new_version_warning:
                click.echo(self.new_version_warning, err=True)
            else:
                logger.info(f'{self.prog_name} : no new version available')  # pylint: disable=logging-fstring-interpolation
        except Exception as error:  # pylint: disable=broad-except
            logger.exception(error)



'''Preconfigured version command for root AdvancedGroup (skeleton)'''
import click
from click_help_colors import HelpColorsCommand  # type: ignore


@click.command(short_help='Print version', cls=HelpColorsCommand, help_headers_color='yellow', help_options_color='green')  # type: ignore
@click.pass_context
def version_cmd(ctx: click.Context) -> None:
    '''Print version, equivalent to -V and --version'''
    click.echo(f"{click.style(ctx.obj.prog_name, fg='yellow')}, version {click.style(ctx.obj.version, fg='green')}", color=ctx.color)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['click_skeleton']

package_data = \
{'': ['*']}

install_requires = \
['click-aliases>=1.0.1,<2.0.0',
 'click-completion>=0.5.0,<0.6.0',
 'click-didyoumean>=0.3.0,<0.4.0',
 'click-help-colors>=0.9,<0.10',
 'click-option-group>=0.5.1,<0.6.0',
 'click>=8.0.1,<9.0.0',
 'munch>=2.5.0,<3.0.0',
 'requests>=2.24.0,<3.0.0',
 'semver>=2.10.2,<3.0.0',
 'types-requests>=2.25.0,<3.0.0']

setup_kwargs = {
    'name': 'click-skeleton',
    'version': '0.20',
    'description': 'Click app skeleton',
    'long_description': None,
    'author': 'Adrien Pensart',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)



'''Improved click.testing test runner'''
import logging
import traceback
from typing import Any
import click
from click import testing
from click_skeleton import helpers

logger = logging.getLogger(__name__)


def run_cli(cli_runner: testing.CliRunner, called_cli: click.core.Group, *args: Any, **kwargs: Any) -> str:
    '''Helper to run the CLI for pytest-click unit tests'''
    result: testing.Result = cli_runner.invoke(
        called_cli,
        *args,
        **kwargs,
    )
    logger.debug(result.output)
    if result.exception:
        exc_info = result.exc_info
        traceback.print_exception(*exc_info)  # type: ignore
    if result.exit_code != 0:
        elems = helpers.recursive_str(*args)
        elems = ' '.join(elems)
        logger.error(f'Failed : {cli_runner.get_default_prog_name(called_cli)} {elems}')  # pylint:disable=logging-fstring-interpolation
    assert result.exit_code == 0
    return result.output.rstrip()



'''Defaults settings'''
DEFAULT_CONTEXT_SETTINGS = {
    'max_content_width': 140,
    'terminal_width': 140,
    'help_option_names': ['-h', '--help'],
    # 'help_headers_color': 'yellow',
    # 'help_options_color': 'green',
}
